# -*- coding: UTF-8 -*-
from datetime import date, time, datetime
from decorated.base.context import Context
from json.encoder import JSONEncoder
from loggingd import log_enter, log_error
from metaweb import views
from metaweb.errors import WebError
from metaweb.resps import Response, NotFoundResponse
import doctest
import json
import logging

log = logging.getLogger(__name__)

class JsonEncoder(JSONEncoder):
    def default(self, obj):
        if hasattr(obj, '__json__'):
            return getattr(obj, '__json__')()
        elif isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(obj, date):
            return obj.strftime('%Y-%m-%d')
        elif isinstance(obj, time):
            return obj.strftime('%H:%M:%S')
        else:
            return obj
    
def coor_maker(base_class=object, context_class=None):
    context_class_ = context_class or Context
    class _Coor(base_class):
        context_class = context_class_
        
        @log_enter('Handling url {path} ...')
        @log_error('Failed to handle url {path}.', exc_info=True)
        def render(self, path):
            try:
                path = _match_view(path)
            except Response as resp:
                return resp
            fields, headers, cookies = self._parse_request()
            request = {
                'cookies': cookies,
                'fields': fields,
                'headers': headers,
                'path': path,
            }
            try:
                with self.context_class(request=request):
                    result = path.view.render(request)
                    resp = _encode_result(result, path.view.mimetype, headers.get('Accept')) # need context
            except Response as resp:
                pass
            except Exception as e:
                log.warning('Failed to render.', exc_info=True)
                if not isinstance(e, WebError):
                    e = WebError(500, 'INTERNAL_ERROR', 'Server encountered an internal error.')
                resp = _encode_result(e, path.view.mimetype, headers.get('Accept'))
            return resp
        
        def _parse_request(self):
            fields = self._read_fields()
            headers = self._read_headers()
            cookies = self._read_cookies()
            for k, v in fields.items():
                log.debug('Field %s=%s' % (k, v))
            return fields, headers, cookies
            
        def _read_cookies(self):
            return {}

        def _read_fields(self):
            raise NotImplementedError()
        
        def _read_headers(self):
            return {}
    return _Coor

def _encode_result(resp, mimetype, accept):
    '''
    >>> resp = _encode_result(Response(200, 'aaa'), 'application/octet-stream', None)
    >>> resp.body
    'aaa'
    
    >>> resp = _encode_result('\xff\xff', 'application/octet-stream', None)
    >>> resp.status
    200
    >>> resp.body
    '\\xff\\xff'
    >>> resp.headers['Content-Type']
    'application/octet-stream'
    
    >>> resp = _encode_result('aaa', 'text/html; charset=utf-8', None)
    >>> resp.status
    200
    >>> resp.body
    'aaa'
    >>> resp.headers['Content-Type']
    'text/html; charset=utf-8'
    
    >>> resp = _encode_result('aaa', 'text/html', 'application/json')
    >>> resp.body
    '"aaa"'
    >>> resp.headers['Content-Type']
    'application/json'
    
    >>> resp = _encode_result(WebError(400, 'ERROR_CODE', 'Error message.'), 'application/octet-stream', None)
    >>> resp.status
    400
    >>> resp.body
    '[ERROR_CODE] Error message.'
    >>> resp.headers['Content-Type']
    'application/octet-stream'
    
    >>> resp = _encode_result(WebError(400, 'ERROR_CODE', 'Error message.'), 'application/octet-stream', 'application/json')
    >>> resp.status
    400
    >>> resp.body
    '{"message": "Error message.", "code": "ERROR_CODE"}'
    >>> resp.headers['Content-Type']
    'application/json'
    '''
    if isinstance(resp, Response):
        return resp
    else:
        status = resp.status if isinstance(resp, WebError) else 200
        if accept is not None:
            mimetype = accept
        if mimetype == 'application/json':
            resp = json.dumps(resp, cls=JsonEncoder)
            mimetype = 'application/json'
        elif mimetype.startswith('text/') or isinstance(resp, WebError):
            resp = unicode(resp).encode('utf-8')
        headers = {'Content-Type': mimetype}
        return Response(status, resp, headers=headers)

@log_error('View for {path} cannot be found.')
def _match_view(path):
    path = views.match(path)
    if path is None:
        raise NotFoundResponse()
    return path

if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: UTF-8 -*-
from decorated import Function
from decorated.base.function import ArgError
from decorated.util import modutil
from metaweb.errors import ValidationError
from metaweb.path import Path
from metaweb.resps import RedirectResponse
import doctest
import loggingd
import re

REGEX_PATH_PARAM = re.compile('<(.*?)(:.*?)?>')
log = loggingd.getLogger(__name__)
_pending_views = []
_abs_pathes = {}
_regex_pathes = {}

class View(Function):
    mimetype = 'application/octet-stream'
    
    def bind(self, path):
        self.path = path
        if '<' in path and '>' in path:
            for name, type in REGEX_PATH_PARAM.findall(path):
                if type == '':
                    type = unicode
                elif type == ':int':
                    type = int
                else:
                    raise Exception('Unknown type "%s" in path "%s".' % (type[1:], path))
                self._path_params[name] = type
            path = REGEX_PATH_PARAM.sub(lambda m: '(?P<%s>.*?)' % m.group(1), path)
            path = re.compile('^%s$' % path)
            _regex_pathes[path] = self
        else:
            _abs_pathes[path] = self
    
    def render(self, request):
        args = self._parse_args(request['fields'], request['path'].args)
        args = self._resolve_args(**args)
        return self._call(**args)
        
    def _decorate(self, func):
        super(View, self)._decorate(func)
        _pending_views.append(self)
        return self
    
    def _init(self, path=None, mimetype=None):
        super(View, self)._init()
        self._specified_path = path
        if mimetype is not None:
            self.mimetype = mimetype
        self._path_params = {}
        
    def _parse_args(self, fields, path_args):
        results = {}
        defaults = dict(self.optional_params)
        for param in self.params:
            if param in fields:
                results[param] = fields[param]
            elif param in path_args:
                type = self._path_params[param]
                results[param] = type(path_args[param])
            elif param in defaults:
                results[param] = defaults[param]
            else:
                raise ValidationError('ARGUMENT_MISSING', param=param)
        return results
    
class Api(View):
    mimetype = 'application/json'
    
class Page(View):
    mimetype = 'text/html; charset=utf-8'
    
def add_default_view(url):
    @View
    def _default():
        raise RedirectResponse(url)
    _default.bind('/')
    
def load(roots=('views',)):
    if isinstance(roots, (list, tuple)):
        roots = {r: '' for r in roots}
    for root in roots:
        modutil.load_tree(root)
    for v in _pending_views:
        obj_path = _obj_to_path(v)
        path = _calc_path(roots, obj_path, v._specified_path)
        if path is not None:
            v.bind(path)
        
def match(path):
    view = _abs_pathes.get(path)
    if view is not None:
        path = Path(path)
        path.view = view
        return path
    for pattern, view in _regex_pathes.items():
        match = pattern.match(path)
        if match is None:
            continue
        args = match.groupdict()
        path = Path(path)
        path.view = view
        path.args = args
        return path
    return None

def _calc_path(roots, obj_path, specified_path):
    '''
    >>> _calc_path({'views': ''}, 'views.users.get', None)
    '/users/get'
    >>> _calc_path({'views': ''}, 'views.get', None)
    '/get'
    >>> _calc_path({'views': '', 'views2': ''}, 'views2.users.get', None)
    '/users/get'
    >>> _calc_path({'admin': '/admin'}, 'admin.users.get', None)
    '/admin/users/get'
    >>> _calc_path({'views': ''}, 'views2.users.get', None) is None
    True
    '''
    for package, prefix in roots.items():
        if obj_path.startswith(package + '.'):
            if specified_path is None:
                specified_path = obj_path[len(package):].replace('.', '/')
            return prefix + specified_path
    return None

def _obj_to_path(obj):
    '''
    >>> from metaweb.views import View
    >>> _obj_to_path(View)
    'metaweb.views.View'
    '''
    return obj.__module__ + '.' + obj.__name__

if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
from decorated.util import modutil
import json

if modutil.module_exists('web'):
    from metaweb import files, coor, views
    from metaweb.files import FileField
    from web import webapi
    import web
    
    class WebpyFileField(FileField):
        def __init__(self, storage):
            super(WebpyFileField, self).__init__(storage.filename)
            self._file = storage.file
            
        def chunks(self, size=files.CHUNK_SIZE):
            while True:
                chunk = self._file.read(size)
                if len(chunk):
                    yield chunk
                else:
                    break
                
    def run(cls=None, context_class=None, default_url=None, roots=('views',)):
        cls = cls or _create_coor(context_class=context_class)
        app = _create_app(cls, default_url, roots)
        app.run()
        
    def wsgi(cls=None, context_class=None, default_url=None, roots=('views',)):
        cls = cls or _create_coor(context_class=context_class)
        app = _create_app(cls, default_url, roots)
        return app.wsgifunc()
        
    def _create_app(cls, default_url, roots):
        views.load(roots=roots)
        if default_url is not None:
            views.add_default_view(default_url)
        mapping = ('(.+)', cls)
        web.config.debug = False
        return web.application(mapping)
        
    def _create_coor(context_class):
        class _Coor(coor.coor_maker(context_class=context_class)):
            def GET(self, path):
                resp = self.render(path)
                return self._process_resp(resp)
        
            def POST(self, path):
                resp = self.render(path)
                return self._process_resp(resp)
            
            def _read_cookies(self):
                return web.cookies()
            
            def _read_fields(self):
                env = webapi.ctx.env
                json_fields = {}
                method = env.get('REQUEST_METHOD', '').upper()
                ctype = env.get('CONTENT_TYPE', '').lower()
                if method == 'POST' and ctype.startswith('application/json'):
                    data = env['wsgi.input'].read().decode('utf-8')
                    json_fields = json.loads(data)
                fields = webapi.rawinput(method='both')
                results = {k: v[-1] if isinstance(v, list) else v for k, v in fields.items()}
                results = {k: v.decode('utf-8') for k, v in results.items()}
                results.update(json_fields)
                return results
            
            def _read_headers(self):
                def _normalize_key(key):
                    key = key[len('HTTP_'):]
                    ss = key.split('_')
                    ss = [s.capitalize() for s in ss]
                    return '-'.join(ss)
                return {_normalize_key(k): v \
                        for k, v in web.ctx.env.items() \
                        if k.startswith('HTTP_')}
            
            def _process_resp(self, resp):
                web.ctx.status = resp.status_string()
                for k, v in resp.headers.items():
                    web.header(k, str(v))
                return resp.body
        return _Coor
